import asyncio
from urllib.parse import urlsplit
from typing import Dict, Union, Tuple
from time import time
import random


class RateLimiter:
    """Utility class to compute appropriate amount of time to sleep based on a host's specified rate limit and request history to this host."""

    def __init__(self,
                 # rate limit to use for all hosts that don't specify a rate limit.
                 default_rate_limit: Union[float, Tuple[float]] = None,
                 # map host names to rate limit in requests per second.
                 host_requests_per_sec: Dict[str, Union[float, Tuple[float]]] = None):
        self.default_rate_limit = default_rate_limit
        self._host_requests_per_sec: Dict[str, float] = {}
        if host_requests_per_sec is not None:
            self.set_host_requests_per_sec(host_requests_per_sec)
        # map host name to time that last request was dispatched to this host.
        self._host_last_request_time: Dict[str, int] = {}

    @property
    def host_requests_per_sec(self) -> Dict[str, Union[float, Tuple[float]]]:
        """Return map of all hostnames to non-default rate limits."""
        return {host: l['limit'] for host, l in self._host_requests_per_sec.items() if l['limit'] != self.default_rate_limit}

    def set_host_requests_per_sec(self, host_requests_per_sec: Dict[str, Union[float, Tuple[float]]]) -> None:
        """Map host name to max requests per second for that host."""
        for host, limit in host_requests_per_sec.items():
            # check that any limit tuple has two elements (lower limit, upper limit)
            if isinstance(limit, (tuple, list)) and len(limit) != 2:
                raise ValueError(
                    f'Rate limit tuple should have 2 elements! (lower limit, upper limit). Provided limit has {len(limit)} elements: {limit}')
            self._host_requests_per_sec[host] = {
                'limit': limit, 'lock': asyncio.Lock()}

    async def maybe_sleep(self, url: str) -> None:
        """Sleep if this URL's host has a rate limit that has been reached."""
        host = urlsplit(url).netloc
        # check if a specific rate limit has been assigned to this host.
        if host not in self._host_requests_per_sec:
            if self.default_rate_limit is None:
                # host does not have an assigned rate limit and there is no default. no need to sleep.
                return
            # host does not have an assigned rate limit, so host should use the default limit.
            self._host_requests_per_sec[host] = {
                'limit': self.default_rate_limit, 'lock': asyncio.Lock()}
        async with self._host_requests_per_sec[host]['lock']:
            # sleep to not exceed max requests per second.
            await asyncio.sleep(self._sleep_time(host))
        # record time that request is being dispatched.
        self._host_last_request_time[host] = time()

    def _sleep_time(self, host: str) -> float:
        """Determine appropriate time to sleep (if any) based on rate limit for this host and request history to this host."""
        if host not in self._host_last_request_time:
            # no previous request has been made to this host, so no need to sleep.
            return 0.0
        rate_limit = self._host_requests_per_sec[host]['limit']
        if isinstance(rate_limit, (tuple, list)):
            rate_limit = random.uniform(*rate_limit)
        # s/req - time since last request
        sleep_time = (1 / rate_limit) - (time() -
                                         self._host_last_request_time[host])
        return max(0, sleep_time)
